/*{{{}}}*/
/*{{{  #include*/
#include <stdlib.h>
#include <math.h>
#include "fft.h"
/*}}}  */
/*{{{  #define*/
#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif
/*}}}  */
/*{{{  sintbl(n, sinp)*/
void
sintbl(n, sinp)
    int             n;
    double          *sinp;
{
    int             i, n2, n4, n8;
    double          c, s, dc, ds, t;

    n2 = n / 2;
    n4 = n / 4;
    n8 = n / 8;
    t = sin(M_PI / (double)n);
    dc = 2.0 * t * t;
    ds = sqrt(dc * (2.0 - dc));
    t = 2.0 * dc;
    c = sinp[n4] = 1.0;
    s = sinp[0] = 0;
    for (i = 1; i < n8; i++) {
        c -= dc;
        dc += t * c;
        s += ds;
        ds -= t * s;
        sinp[i] = s;
        sinp[n4 - i] = c;
    }

    if (n8 != 0){
        sinp[n8] = sqrt(0.5);
    }
    for (i = 0; i < n4; i++){
        sinp[n2 - i] = sinp[i];
    }
    for (i = 0; i < n2 + n4; i++){
        sinp[i + n2] = -sinp[i];
    }
}
/*}}}  */
/*{{{  bitrev(n, bitrevp)*/
void
bitrev(n, bitrevp)
    int             n;
    int            *bitrevp;
{
    int             i, j, k, n2;

    n2 = n / 2;
    i = j = 0;
    for (;;) {
        bitrevp[i] = j;
        if (++i >= n){
            break;
        }
        k = n2;
        while (k <= j) {
            j -= k;
            k /= 2;
        }
        j += k;
    }
}
/*}}}  */
/*{{{  pre_fft(n, dir)*/
fft_plan *
pre_fft(n, dir)
    int n;
    int dir;
{
    static fft_plan p;
    static int used = 0;

    if(n != 0){
        /*{{{  create fft plan*/
        if(n <= 1){
            return(NULL);
        }else{
            p.dir = dir;
            p.n = n;
            if(used == 1){
                free(p.sin);
                free(p.rev);
            }
            p.sin = malloc((n + n/4) * sizeof(double));
            if(p.sin == NULL){
                return(NULL);
            }
            p.rev = malloc( n        * sizeof(int   ));
            if(p.rev == NULL){
                free(p.sin);
                return(NULL);
            }
            sintbl(n, p.sin);
            bitrev(n, p.rev);
            used = 1;
            return(&p);
        }
        /*}}}  */
    }else{
        /*{{{  destroy fft plan*/
        if(used == 1){
            free(p.sin);
            free(p.rev);
        }
        used = 0;
        return(NULL);
        /*}}}  */
    }
}
/*}}}  */
/*{{{  fft(p, re, im)*/
void
fft(p, re, im)
    /*{{{  variables*/
    fft_plan       *p;
    double         *re;
    double         *im;
    /*}}}  */
{
    /*{{{  variables*/
    register int    i, j, k, n, ik;
    register double c, s, t, u, v, w;
    int             n4, h, d, k2;
    int             dir;
    double          *sinp;
    int             *revp;
    /*}}}  */
    /*{{{  variable preparations*/
    n = p->n;
    n4 = p->n / 4;
    sinp = p->sin;
    revp = p->rev;
    dir = (p->dir == FFT_BACKWARD) ? -1 : 1;
    /*}}}  */

    for (i = 0; i < n; i++) {
        j = revp[i];
        if (i < j) {
            t = re[i];
            u = re[j];
            re[i] = u;
            re[j] = t;
            t = im[i];
            u = im[j];
            im[i] = u;
            im[j] = t;
        }
    }
    for (k = 1; k < n; k = k2) {
        h = 0;
        k2 = k + k;
        d = n / k2;
        for (j = 0; j < k; j++) {
            c = sinp[h + n4];
            s = dir * sinp[h];
            for (i = j; i < n; i += k2) {
                ik = i + k;
                v = re[ik];
                w = re[ik];
                u = im[ik];
                t = im[ik];
                t *= s;
                u *= c;
                v *= c;
                w *= s;
                t += v;
                u -= w;
                im[ik] = im[i] - u;
                im[i] += u;
                re[ik] = re[i] - t;
                re[i] += t;
            }
            h += d;
        }
    }
    /*{{{  backward FFT scaling*/
    if(dir == FFT_BACKWARD){
        s = 1.0 / (double)n;
        for(i = 0; i < n;i+=2){
             re[i] *= s;
             re[i+1] *= s;
             im[i] *= s;
             im[i+1] *= s;
        }
    }
    /*}}}  */
}
/*}}}  */
